/** @file addidata_console.h
* @brief console related utility functions
*
* @author Sylvain Nahas
* @date Jan 27, 2009
*
* @version  $LastChangedRevision: 3021 $
* @version $LastChangedDate: 2008-06-11 10:16:10 +0200 (Wed, 11 Jun 2008) $
*
*
* @par LICENCE
* @verbatim
   Copyright (C) 2009  ADDI-DATA GmbH for the source code of this module.

        ADDI-DATA GmbH
        Airport boulevard B210
        D-77836 Rheinmuenster
        Tel: +49(0)7229/1847-0
        Fax: +49(0)7229/1847-200
        http://www.addi-data-com
        info@addi-data.com

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   You shoud also find the complete GPL in the COPYING file
   accompanying this source code.
   @endverbatim
*
*
*/


#ifndef ADDIDATA_CONSOLE_H_
#define ADDIDATA_CONSOLE_H_

#ifdef WIN32
	#include <stdint_compat.h>
	#define __INLINE__ __inline
	#include <windows.h>
	typedef HANDLE key_thread_handle_t;
#else
	#include <stdint.h>
	#include <pthread.h>
	#define __INLINE__ inline
	typedef pthread_t * key_thread_handle_t;
#endif

/** require the user to enter an integer from stdin
 * in the range "min" to "max" included
 *
 * the convention is signed decimal
 *
 * @param prompt a string to print - may be NULL
 * @param min begin of the legal range
 * @param max end of the legal range
 *
 * @retval the integer entered by the user.
 *
 * @note this function handles parse errors
 * and loop until a legal integer has been entered
 */
int addidata_console_get_integer(const char *prompt, int min, int max);

/** require the user to enter an unsigned integer from stdin
 * in the range "min" to "max" included
 *
 * the convention is unsigned decimal
 *
 * @param prompt a string to print - may be NULL
 * @param min begin of the legal range
 * @param max end of the legal range
 *
 * @retval the unsigned integer entered by the user.
 *
 * @note this function handles parse errors
 * and loop until a legal integer has been entered
 */
unsigned int addidata_console_get_unsigned_integer(const char *prompt, unsigned int min, unsigned int max);

/** require the user to enter an integer from stdin
 * in the range "min" to "max" included
 *
 * the convention is hexadecimal
 *
 * @param prompt a string to print - may be NULL
 * @param min begin of the legal range
 * @param max end of the legal range
 *
 * @retval the integer entered by the user.
 *
 * @note this function handles parse errors
 * and loop until a legal integer has been entered
 */
int addidata_console_get_hexa(const char *prompt, int min, int max);

/** require the user to enter an unsigned integer from stdin
 * in the range "min" to "max" included
 *
 * the convention is unsigned hexadecimal
 *
 * @param prompt a string to print - may be NULL
 * @param min begin of the legal range
 * @param max end of the legal range
 *
 * @retval the unsigned integer entered by the user.
 *
 * @note this function handles parse errors
 * and loop until a legal integer has been entered
 */
unsigned int addidata_console_get_unsigned_hexa(const char *prompt, unsigned int min, unsigned int max);

/** print the prompt and expect 0 or 1
 *
 */
int addidata_console_get_yn(char* prompt);

/** simple function to wait for a character to be entered on stdin.
 * may be used to not stop a program automatically.
 *
 * The input character is threashed.
 */
void addidata_console_wait_for_input(void);

/** simple function to wait for a character to be entered on stdin.
 * may be used to not stop a program automatically.
 *
 * @retval The character read.
 */
int addidata_console_wait_for_character(void);

/** Display a timeval struct as human friendly string
*
* Convert date and time element
* of a timeval struct as
* day/month/years hour:min.:sec.:msec.:usec.
*
* @retval >0 Upon successful return, these functions return the number of characters printed (not including the trailing null character used to end output to strings).
* @retval <0 If an output error is encountered, a negative value is returned.
*/
int addidata_fprintf_timestamp(FILE* stream, const char* termfmt, uint32_t seconds, uint32_t useconds);

//---------------------------------------------------------------------------------
typedef struct
{
	int character;
	unsigned int *flag;
} wait_key_input_thread_parameters_t;
key_thread_handle_t addidata_start_key_thread(wait_key_input_thread_parameters_t *params);
void addidata_stop_key_thread(key_thread_handle_t thandle);

#endif /* ADDIDATA_CONSOLE_H_ */
