using System;
using System.Collections.Generic;
using System.Text;

namespace Monitor
{
    public class Msxexxxx : Msxe
    {
        /// <summary>
        /// System web reference
        /// </summary>
        private MSXExxxxWeb.MSXExxxx sysWeb;

        /// <summary>
        /// Number of acquisition channels
        /// </summary>
        private ulong acquisitionChannelNb = 0;

        /// <summary>
        /// Store the types of the channels
        /// 0: not available
        /// 1: temperature
        /// 2: pressure
        /// 3: transducer
        /// 4: analog input
        /// 5: analog input ICP
        /// 6: digital I/O port
        /// 7: incremental counter
        /// 8: sine/cosine
        /// </summary>
        private List<ulong> channelType;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="ip"></param>
        public Msxexxxx(string ip)
            : base(ip)
        {
            // Nothing specific
        }

        /// <summary>
        /// Override the connect method
        /// </summary>
        /// <param name="error"></param>
        /// <returns>
        /// 0:  success
        /// -1: wrong IP address
        /// -2: system error. Check the error variable for more informations
        /// -3: wrong system type
        /// -4: system not reachable
        /// -5: other error. Check the error variable for more informations
        /// </returns>
        public override int Connect(out string error)
        {
            ulong tmp = 0;

            // Call Connect of the base class
            // If successful, then the connectionState variable is set to 1
            int retval = base.Connect(out error);
            if (retval != 0)
                return retval;

            // Create the web-reference object
            this.sysWeb = new MSXExxxxWeb.MSXExxxx();
            try
            {
                this.sysWeb.Url = "http://" + this.IPAddress + ":5555";
                this.sysWeb.Timeout = 2000;
            }
            catch (Exception) { return -1; }

            // Check if the selected system has some acquisition channels
            retval = this.AcquisitionGetNumberOfChannels(out tmp, out error);
            if (retval != 0)
            {
                this.connectionState = 0;
                return retval;
            }

            // Store number
            this.acquisitionChannelNb = tmp;

            // Any acquisition channel?
            if (tmp == 0)
            {
                return -3;
            }

            // Get the types of the acquisition channels
            this.channelType = new List<ulong>();
            for (ulong i = 0; i < this.acquisitionChannelNb; i++)
            {
                ulong type = 0;
                ulong hwPosition = 0;
                ulong index = 0;
                retval = this.AcquisitionGetChannelInfo(i, out type, out hwPosition, out index, out error);
                if (retval != 0)
                {
                    this.connectionState = 0;
                    return retval;
                }

                this.channelType.Add(type);
            }

            return 0;
        }

        /// <summary>
        /// Get the type (as string) of an acquisition channel
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public String GetAcquisitionChannelType(int index)
        {
            if ((index < 0) || (index >= this.channelType.Count))
                return "Wrong index";

            switch (this.channelType[index])
            {
                case 0:
                    return "Not available";
                case 1:
                    return "Temperature";
                case 2:
                    return "Pressure";
                case 3:
                    return "Transducer";
                case 4:
                    return "Analog input";
                case 5:
                    return "Analog input ICP";
                case 6:
                    return "Digital I/O port";
                case 7:
                    return "Incremental counter";
                case 8:
                    return "Sine/Cosine";
                default:
                    return "Unkwnon";
            }
        }

        /// <summary>
        /// Get the number of acquisition channels
        /// </summary>
        public ulong AcquisitionChannelNb
        {
            get
            {
                return this.acquisitionChannelNb;
            }
        }

        #region "SOAP API"

        /// <summary>
        /// Reads the number of acquisition channels
        /// </summary>
        /// <param name="channelNumber">Number of acquisition channels</param>
        /// <param name="error">Error message set if the return value is -2 or -5</param>
        /// <returns>
        /// 0:  success
        /// -1: you must call the Connect function before
        /// -2: system error. Check the error variable for more informations
        /// -3: the method is not implemented on the system. Please check the type of your system
        /// -4: the system is not reachable (not present on the network, or no SOAP server on it)
        /// -5: other error. Check the error variable for more informations
        /// </returns>
        public int AcquisitionGetNumberOfChannels(out ulong channelNumber, out string error)
        {
            // Default
            channelNumber = 0;
            error = string.Empty;

            // Check the connection state: Connect method should be called first
            if (this.connectionState != 1)
            {
                return -1;
            }

            // Call the SOAP function
            try
            {
                MSXExxxxWeb.DefaultResponse response = this.sysWeb.AcquisitionGetNumberOfChannels(0, out channelNumber);
                if (response.iReturnValue != 0)
                {
                    error = "The system returned an error when calling the MSXExxxx__AcquisitionGetNumberOfChannels function." + Environment.NewLine;
                    error += "Please check the documentation of this function for more information." + Environment.NewLine;
                    error += "(iReturnValue: " + response.iReturnValue.ToString() + ", syserrno: " + response.syserrno.ToString() + ")";
                    return -2;
                }
            }
            catch (System.Web.Services.Protocols.SoapHeaderException)
            {
                return -3;
            }
            catch (System.Net.WebException)
            {
                return -4;
            }
            catch (Exception ex)
            {
                error = "Unhandled exception occured: " + ex.Message;
                return -5;
            }
            return 0;
        }

        /// <summary>
        /// Return the selected acquisition channel type and hardware position. 
        /// </summary>
        /// <param name="channel">Selected acquisition channel </param>
        /// <param name="type">Acquisition channel type </param>
        /// <param name="hwPosition">Hardware position index (0 to 7)</param>
        /// <param name="index"> Return the functionality channel index</param>
        /// <param name="error">Error message set if the return value is -2 or -5</param>
        /// <returns>
        /// 0:  success
        /// -1: you must call the Connect function before
        /// -2: system error. Check the error variable for more informations
        /// -3: the method is not implemented on the system. Please check the type of your system
        /// -4: the system is not reachable (not present on the network, or no SOAP server on it)
        /// -5: other error. Check the error variable for more informations
        /// </returns>
        public int AcquisitionGetChannelInfo(ulong channel, out ulong type, out ulong hwPosition, out ulong index, out string error)
        {
            // Default
            type = 0;
            hwPosition = 0;
            index = 0;
            error = string.Empty;

            // Check the connection state: Connect method should be called first
            if (this.connectionState != 1)
            {
                return -1;
            }

            // Call the SOAP function
            try
            {
                MSXExxxxWeb.DefaultResponse response = this.sysWeb.AcquisitionGetChannelInfo(channel, 0, out type, out hwPosition, out index);
                if (response.iReturnValue != 0)
                {
                    error = "The system returned an error when calling the MSXExxxx__AcquisitionGetChannelInfo function." + Environment.NewLine;
                    error += "Please check the documentation of this function for more information." + Environment.NewLine;
                    error += "(iReturnValue: " + response.iReturnValue.ToString() + ", syserrno: " + response.syserrno.ToString() + ")";
                    return -2;
                }
            }
            catch (System.Web.Services.Protocols.SoapHeaderException)
            {
                return -3;
            }
            catch (System.Net.WebException)
            {
                return -4;
            }
            catch (Exception ex)
            {
                error = "Unhandled exception occured: " + ex.Message;
                return -5;
            }
            return 0;
        }

        /// <summary>
        /// Start a sequence acquisition
        /// </summary>
        /// <param name="channels">List of the channels to use</param>
        /// <param name="error">Error message set if the return value is -2 or -5</param>
        /// <returns>
        /// 0:  success
        /// -1: you must call the Connect function before
        /// -2: system error. Check the error variable for more informations
        /// -3: the method is not implemented on the system. Please check the type of your system
        /// -4: the system is not reachable (not present on the network, or no SOAP server on it)
        /// -5: other error. Check the error variable for more informations
        /// </returns>
        public int AcquisitionSequenceInitAndStart(List<int> channels, out string error)
        {
            // Default
            error = string.Empty;

            // Check the connection state: Connect method should be called first
            if (this.connectionState != 1)
            {
                return -1;
            }

            MSXExxxxWeb.AcquisitionSequenceInitAndStartChannelListParam channelList = new MSXExxxxWeb.AcquisitionSequenceInitAndStartChannelListParam();
            channelList.ulChannelList = new ulong[16];
            for (int i = 0; i < channels.Count; i++)
                channelList.ulChannelList[i] = (ulong)channels[i];

            // Call the SOAP function
            try
            {
                int syserrno = 0;
                int retval = this.sysWeb.AcquisitionSequenceInitAndStart((ulong)channels.Count,
                    channelList,
                    1,
                    2, // second
                    0,
                    1,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0x11,
                    1, // force the start
                    0, 0, 0,
                    out syserrno);
                if (retval != 0)
                {
                    error = "The system returned an error when calling the MSXExxxx__AcquisitionSequenceInitAndStart function." + Environment.NewLine;
                    error += "Please check the documentation of this function for more information." + Environment.NewLine;
                    error += "(iReturnValue: " + retval.ToString() + ", syserrno: " + syserrno.ToString() + ")";
                    return -2;
                }
            }
            catch (System.Web.Services.Protocols.SoapHeaderException)
            {
                return -3;
            }
            catch (System.Net.WebException)
            {
                return -4;
            }
            catch (Exception ex)
            {
                error = "Unhandled exception occured: " + ex.Message;
                return -5;
            }
            return 0;
        }

        /// <summary>
        /// Stop everything (without exit codes)
        /// </summary>
        public void StopAll()
        {
            try
            {
                int syserrno = 0;
                this.sysWeb.AcquisitionSequenceStopAndRelease(out syserrno);
            }
            catch (Exception) { }
            try
            {
                int syserrno = 0;
                this.sysWeb.AcquisitionAutoRefreshStopAndRelease(out syserrno);
            }
            catch (Exception) { }
        }

        #endregion
    }
}