using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace Initialisation
{
    public partial class MainForm : Form
    {
        /// <summary>
        /// Store the MSX-E object
        /// </summary>
        private Msxexxxx msxe;

        /// <summary>
        /// Constructor of the graphical interface
        /// </summary>
        public MainForm()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Event handler: occurs once when the form is loaded
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            // Load default parameters
            this.ipTextBox.Text = Properties.Settings.Default.ip;
        }

        /// <summary>
        /// Event handler: occurs when the form is being closed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            Properties.Settings.Default.ip = this.ipTextBox.Text;
            Properties.Settings.Default.Save();
        }

        /// <summary>
        /// Event handler: click on the "Initialise connection" button
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void initialiseConnectionButton_Click(object sender, EventArgs e)
        {
            // Check the validity of the IP address
            if (this.ipTextBox.Valid == false)
                return;

            // Update some graphics
            this.initialiseConnectionButton.Enabled = false;
            this.actionToolStripProgressBar.Visible = true;
            this.actionToolStripStatusLabel.Text = "Initialise the connection with the MSX-E system.";
            this.actionToolStripStatusLabel.Visible = true;
            
            this.digitalIOGroupBox.Enabled = false;
            this.digitalIODataGridView.Rows.Clear();

            // Start the connection background worker
            this.connectionBackgroundWorker.RunWorkerAsync(this.ipTextBox.Text);
        }

        /// <summary>
        /// Job of the connection background worker
        /// The connection task is done in background since it may takes some time to initialise the system
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void connectionBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            int retval = 0;
            string error = string.Empty;

            // Construct the new MSX-E object
            Msxexxxx m = new Msxexxxx(this.ipTextBox.Text);

            // Do the connection
            retval = m.Connect(out error);
            e.Result = new ConnectionWorkerResult(m, retval, error);
        }

        /// <summary>
        /// Event handler: Connection background worker is completed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void connectionBackgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            ConnectionWorkerResult r = (ConnectionWorkerResult)e.Result;

            // Error ?
            if (r.retval != 0)
            {
                string error = string.Empty;
                this.actionToolStripProgressBar.Visible = false;
                this.actionToolStripStatusLabel.Text = "Error occurred.";
                // Reset the system type
                this.systemInformationTextBox.Text = string.Empty;
                // Display the right text
                switch (r.retval)
                {
                    case -1:
                        error = "The selected IP address is wrong. Please change it and retry.";
                        break;
                    case -2:
                        error = "A system error occured: " + r.error;
                        break;
                    case -3:
                        error = "The selected system cannot be used with this sample.";
                        break;
                    case -4:
                        error = "It is not possible to open a connection with the selected system.";
                        break;
                    default:
                        error = "An error occured: " + r.error;
                        break;
                }
                // Display the error
                MessageBox.Show(error, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                // Store the object
                this.msxe = r.msxe;

                this.actionToolStripProgressBar.Visible = false;
                this.actionToolStripStatusLabel.Text = "Connection opened with the system: " + r.msxe.IPAddress;
                // Display the system type
                this.systemInformationTextBox.Text = r.msxe.SystemType;

                // Enable controls
                this.digitalIOGroupBox.Enabled = true;

                // Set data grid
                for (int i = 0; i < (int)this.msxe.DigIOChannelNb; i++)
                {
                    DigitalIO info = this.msxe.GetDigitalIoInfo(i);
                    int index = digitalIODataGridView.Rows.Add();
                    digitalIODataGridView.Rows[index].Cells[0].Value = i.ToString();
                    digitalIODataGridView.Rows[index].Cells[2].Value = info.Port.ToString();

                    // Set the datasource
                    DataGridViewComboBoxCell cell = new DataGridViewComboBoxCell();
                    cell = (DataGridViewComboBoxCell)digitalIODataGridView.Rows[index].Cells[1];
                    if (info.InputPossible == true)
                    {
                        cell.Items.Add("Input");
                    }
                    if (info.OutputPossible == true)
                    {
                        cell.Items.Add("Output");
                    }
                    if (info.CurrentConfiguration == 0)
                    {
                        cell.Value = "Input";
                    }
                    else
                    {
                        cell.Value = "Output";
                    }
                }
            }

            // Enable the "Initialise Connection" button
            this.initialiseConnectionButton.Enabled = true;
        }

        /// <summary>
        /// Event handler: click on "Apply Configuration" button
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void applyConfigurationButton_Click(object sender, EventArgs e)
        {
            applyConfigurationButton.Enabled = false;
            int index = 0;
            for (int i = 0; i < (int)this.msxe.DigIOPortNb; i++)
            {
                DigitalIOPort port = this.msxe.GetDigitalIoPortInfo(i);
                ulong direction = 0;
                int retval = 0;
                string error  =string.Empty;

                // Construct direction
                for (int j = 0; j < (int)port.ChannelNumber; j++)
                {
                    if ("Output" == (string)digitalIODataGridView.Rows[index].Cells[1].Value)
                    {
                        direction = direction + (ulong)(1 << j);
                    }
                    index++;
                }
                direction = direction & 0xFFFFFFFF;

                // Set configuration
                retval = this.msxe.DigitalIOSetPortDirections((ulong)i, direction, out error);
                if (retval != 0)
                {
                    switch (retval)
                    {
                        case -1:
                            error = "The selected IP address is wrong. Please change it and retry.";
                            break;
                        case -2:
                            error = "A system error occured: " + error;
                            break;
                        case -3:
                            error = "The selected system cannot be used with this sample.";
                            break;
                        case -4:
                            error = "It is not possible to open a connection with the selected system.";
                            break;
                        default:
                            error = "An error occured: " + error;
                            break;
                    }

                    // Display the error
                    MessageBox.Show(error, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    applyConfigurationButton.Enabled = true;
                    return;
                }
            }

            // Display success message
            MessageBox.Show("Configuration successfully applied", 
                "Information", 
                MessageBoxButtons.OK, 
                MessageBoxIcon.Information);

            applyConfigurationButton.Enabled = true;
        }
    }
}